# Google Summer of Code ideas for 2019

**Important note for Summer of Code applicants:** When getting in touch with us,
please understand that we will not reply to generic emails.
We recommend that you read our documentation and perhaps even try following
the [walkthrough](../glsl-fuzz-walkthrough.md) to get a high-level idea of how the project works.
We welcome your questions, ideas, and feedback.

Please also see the [GSoC student guide](https://google.github.io/gsocguides/student/), in
particular [how to make first
contact](https://google.github.io/gsocguides/student/making-first-contact).

We look forward to working with you!

## Improve information logging in graphics driver test harnesses

### Description

The Android Graphics Tools team rely on several test harnesses to run tests on
graphics drivers. Each time a test runs, a test harness typically logs relevant
information such as which driver primitives are called, their execution
duration, etc. It is important to gather relevant information to detect driver
bugs as soon as possible, and to ease subsequent debugging.

This project aims at improving the information logging for the various test
harnesses used by the team, to report relevant information to
help detecting and debug driver mis-behaviours (e.g. crashes, but also
timeouts).

### Expected outcomes

The project will involve improving and unifying information logging the test harnesses associated with:

* [Amber](https://github.com/google/amber)
* [GraphicsFuzz](https://github.com/google/graphicsfuzz) GLES worker
* the [GAPID](https://github.com/google/gapid) replayer

The GAPID replayer is considered as a stretch goal since it is a whole virtual
machine in itself.

### Required skills

* C/C++ and Python programming
* Operating systems basics: process management, multithreading

Experience with drivers or graphics APIs such as OpenGL or Vulkan would be
useful but is not required.

### Possible mentors

Hugues Evrard (primary), Paul Thomson (secondary)

### Difficulty rating

Easy.


## Improve shading language support in GraphicsFuzz

### Description

The [GraphicsFuzz](https://github.com/google/graphicsfuzz) tool finds defects in graphics shader compilers and other shader processing tools by generating and transforming shaders expressed in the OpenGL shading language (GLSL).  The tool currently supports a reasonable subset of GLSL, but GLSL is a rich language with a large set of built-in functions, many of which GraphicsFuzz does not yet know how to manipulate.

This project will involve teaching GraphicsFuzz about the semantics of a much larger set of GLSL features, and devising new program transformations related to these features in order to find a wider range of bugs.

### Expected outcomes

By the end of the project, GraphicsFuzz should be augmented to fully support (at a basic level) the built-in functions from the OpenGL ES 1.00, 3.00, 3.10 and 3.20 shading languages.  In addition, at least a handful of new transformations involving features of these shading languages should have been implemented.

The project is open-ended; a stretch goal could be to apply these new capabilities to the finding of bugs in open source graphics drivers.

### Required skills

* Java and Python programming

Experience with drivers or graphics APIs such as OpenGL or Vulkan would be
useful but is not required.

### Possible mentors

Alastair Donaldson (primary), Paul Thomson (secondary)

### Difficulty rating

Easy, with high potential for open-ended, more challenging, extensions.


## Web-based visualization of SPIR-V

See the [Getting started](2019-vis.md) page for this project.

### Description

[SPIR-V](https://www.khronos.org/registry/spir-v/) is an intermediate representation for programming GPUs, along the lines
of LLVM IR.

When debugging SPIR-V code or tools that operate on SPIR-V, the IR can be hard
to understand.

This project is about creating a visualization tool for SPIR-V programs,
inspired by [what the Go compiler can
do](https://pauladamsmith.com/blog/2016/08/go-1.7-ssa.html):

![Go visualization example](https://pauladamsmith.com/images/gossa/thumb.png)

### Expected outcomes

Possible features of the tool could include:

* Showing the control flow graph (i.e. nodes and edges), with some informative highlighting of the
  instructions inside nodes

* Interactive features, such as moving CFG nodes around, or clicking on an ID to
  highlight its uses

* Providing a diff view to see the effects of program
  optimizations/transformations side-by-side (like the optimizations of the Go
  compiler in the article linked above)

### Required skills

* Familiarity with HTML, JavaScript/TypeScript and/or other technologies for creating the "GUI" viewer page/app.
* A small amount of C++ programming may be needed, depending on your approach; the existing tools for working with SPIR-V are written in C++. See the [getting started](2019-vis.md) page for this project to learn about the different approaches.

Experience with compilers and thinking about control flow graphs (or just graphs) would be
useful but is not required.


### Possible mentors

Paul Thomson (primary), Hugues Evrard (secondary)

### Difficulty rating

Medium.



## Comparative code coverage for shader compilers

### Description

When a program crashes, the stacktrace can provide a good indication of where to
look in the source code to investigate the crash root cause. But when a program
terminates while returning a wrong result, it is often unclear where to look in
the source to known where the issue comes from.

This project aims at using code coverage to provide guidance about where to look
in the source code to investigate shader compilers mis-compilations.

The [gcov](https://gcc.gnu.org/onlinedocs/gcc/Gcov.html) tool reports code
coverage of a given execution of a program. The
[lcov](https://github.com/linux-test-project/lcov) tool can present the coverage
data in a web-based report:

![Lcov example](http://ltp.sourceforge.net/coverage/lcov/screenshot2.gif)

Two executions of the same program, with different inputs, will lead to two
separate coverage reports. In the context of testing shader compilers (part of
GPU drivers), we can compare the shader compiler coverage difference when
compiling two almost-identical shaders, with one of the shaders leading to a
compilation issue.

### Expected outcomes

By the end of the project, we hope the student will have:
1. Extended lcov to generate a diff view between two code coverage reports of the same
   program.
2. Explored how the coverage difference can guide to indicate where to look in
   the shader compiler source code to investigate the root cause of
   mis-compilations.

The second goal will involve working with open-source shader compilers like
[Mesa](https://www.mesa3d.org/) or
[SPIRV-tools](https://github.com/KhronosGroup/SPIRV-Tools).

### Required skills

* C/C++ programming to understand shader compiler implementations
* Interest in compilers
* Basic HTML and CSS knowledge for lcov extension

Experience with gcov/lcov, shader compilers would be useful but is not required.

### Possible mentors

Hugues Evrard (primary), Alastair Donaldson (secondary)

### Difficulty rating

Medium.


## Coverage-guided fuzzing for shader tools

### Description

[GraphicsFuzz](https://github.com/google/graphicsfuzz) is an automated
fuzz-testing framework for tools that operate on graphics shaders, such as
graphics drivers and graphics shader compilers.  Currently GraphicsFuzz is *black-box*:
inputs to the tool under test are generated without any information about
what code is being executed.

This project will investigate extending the GraphicsFuzz approach with
coverage-guided features, where feedback on the code executed by the tool
under test is used to inform the fuzzing process.  This has been shown to be
useful in fuzzers such as [AFL](http://lcamtuf.coredump.cx/afl/).

### Expected outcomes

The core outcome will be a coverage-guided extension to the GraphicsFuzz testing framework that can be applied to open source tools that
operate on shaders, such as:

* [Mesa](https://www.mesa3d.org/) open source graphics drivers
* The [SPIRV-Tools](https://github.com/KhronosGroup/SPIRV-Tools) collection of tools
* The [SwiftShader](https://github.com/google/swiftshader) software renderer

A stretch goal will be to use binary instrumentation to enable coverage-guided
fuzzing for closed-source graphics drivers in Android devices.

### Required skills

* C++ and Java programming
* Tweaking build systems such as CMake

Experience with the asan/msan sanitizers, shader compilers and binary instrumentation would be useful but is not required.


### Possible mentors

Alastair Donaldson (primary), Paul Thomson (secondary)

### Difficulty rating

Medium/hard: the core goal of applying coverage-guided fuzzing to open source tools is medium difficulty; the stretch goal is hard.
